package com.audio.device.manager;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.media.AudioDeviceInfo;
import android.media.AudioManager;
import android.util.Log;
import com.facebook.react.bridge.Arguments;
import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;
import com.facebook.react.bridge.WritableMap;
import java.util.HashSet;
import java.util.Set;

public class RNAudioDeviceModule extends ReactContextBaseJavaModule {
  private static final String REACT_NATIVE_MODULE_NAME = "AudioDeviceManager";
  private static final String TAG = REACT_NATIVE_MODULE_NAME;
  private String mPackageName = "com.audio.device.manager";

  private final ReactApplicationContext reactContext;

  // --- AudioRouteManager
  private AudioManager audioManager;
  private boolean audioManagerActivated = false;
  private boolean isOrigAudioSetupStored = false;
  private boolean origIsSpeakerPhoneOn = false;
  private boolean origIsMicrophoneMute = false;
  private int origAudioMode = AudioManager.MODE_INVALID;
  private boolean defaultSpeakerOn = false;
  private int forceSpeakerOn = 0;
  private static final String ACTION_HEADSET_PLUG = (
      android.os.Build.VERSION.SDK_INT >= 21
    )
    ? AudioManager.ACTION_HEADSET_PLUG
    : Intent.ACTION_HEADSET_PLUG;

  private BroadcastReceiver wiredHeadsetReceiver;

  /**
   * AudioDevice is the names of possible audio devices that we currently
   * support.
   */
  public enum AudioDevice {
    SPEAKER_PHONE,
    WIRED_HEADSET,
    EARPIECE,
    BLUETOOTH,
    NONE,
  }

  /** AudioManager state. */
  public enum AudioManagerState {
    UNINITIALIZED,
    PREINITIALIZED,
    RUNNING,
  }

  private int savedAudioMode = AudioManager.MODE_INVALID;
  private boolean savedIsSpeakerPhoneOn = false;
  private boolean savedIsMicrophoneMute = false;
  private boolean hasWiredHeadset = false;

  // Default audio device; speaker phone for video calls or earpiece for audio
  // only calls.
  private AudioDevice defaultAudioDevice = AudioDevice.NONE;

  // Contains the currently selected audio device.
  // This device is changed automatically using a certain scheme where e.g.
  // a wired headset "wins" over speaker phone. It is also possible for a
  // user to explicitly select a device (and overrid any predefined scheme).
  // See |userSelectedAudioDevice| for details.
  private AudioDevice selectedAudioDevice;

  // Contains the user-selected audio device which overrides the predefined
  // selection scheme.
  // TODO(henrika): always set to AudioDevice.NONE today. Add support for
  // explicit selection based on choice by userSelectedAudioDevice.
  private AudioDevice userSelectedAudioDevice;

  // Contains speakerphone setting: auto, true or false
  private final String useSpeakerphone = SPEAKERPHONE_AUTO;

  // Handles all tasks related to Bluetooth headset devices.
  private final BluetoothManager bluetoothManager;

  @Override
  public String getName() {
    return REACT_NATIVE_MODULE_NAME;
  }

  public RNAudioDeviceModule(ReactApplicationContext reactContext) {
    super(reactContext);
    this.reactContext = reactContext;
    reactContext.addLifecycleEventListener(this);
    audioManager =
      ((AudioManager) reactContext.getSystemService(Context.AUDIO_SERVICE));
    bluetoothManager = BluetoothManager.create(reactContext, this);
  }

  // Get connected audio output devices
  @ReactMethod
  public void getAudioOutputDevices(Promise promise) {
    final AudioDeviceInfo[] devices = audioManager.getDevices(
      AudioManager.GET_DEVICES_ALL
    );

    WritableMap data = Arguments.createMap();

    for (AudioDeviceInfo device : devices) {
      final int type = device.getType();
      if (type == AudioDeviceInfo.TYPE_WIRED_HEADSET) {
        data.putString(
          String.valueOf(AudioDevice.WIRED_HEADSET),
          String.valueOf(device.getId())
        );
      } else if (type == AudioDeviceInfo.TYPE_BLUETOOTH_SCO) {
        data.putString(
          String.valueOf(AudioDevice.BLUETOOTH),
          String.valueOf(device.getId())
        );
      } else if (type == AudioDeviceInfo.TYPE_BUILTIN_EARPIECE) {
        data.putString(
          String.valueOf(AudioDevice.EARPIECE),
          String.valueOf(device.getId())
        );
      }
    }

    promise.resolve(data);
  }

  // Before start to use this service, it's storing original audio setup
  private void storeOriginalAudioSetup() {
    Log.d(TAG, "storeOriginalAudioSetup()");
    if (!isOrigAudioSetupStored) {
      origAudioMode = audioManager.getMode();
      origIsSpeakerPhoneOn = audioManager.isSpeakerphoneOn();
      origIsMicrophoneMute = audioManager.isMicrophoneMute();
      isOrigAudioSetupStored = true;
    }
  }

  // When stop to use this service, restore original audio setup
  private void restoreOriginalAudioSetup() {
    Log.d(TAG, "restoreOriginalAudioSetup()");
    if (isOrigAudioSetupStored) {
      setSpeakerphoneOn(origIsSpeakerPhoneOn);
      setMicrophoneMute(origIsMicrophoneMute);
      audioManager.setMode(origAudioMode);
      if (getCurrentActivity() != null) {
        getCurrentActivity()
          .setVolumeControlStream(AudioManager.USE_DEFAULT_STREAM_TYPE);
      }
      isOrigAudioSetupStored = false;
    }
  }

  // Start to get wiredHeadset event. It's receiving events that wiredheadset plugin
  private void startWiredHeadsetEvent() {
    if (wiredHeadsetReceiver == null) {
      Log.d(TAG, "startWiredHeadsetEvent()");
      IntentFilter filter = new IntentFilter(ACTION_HEADSET_PLUG);
      wiredHeadsetReceiver =
        new BroadcastReceiver() {

          @Override
          public void onReceive(Context context, Intent intent) {
            if (ACTION_HEADSET_PLUG.equals(intent.getAction())) {
              hasWiredHeadset = intent.getIntExtra("state", 0) == 1;
              updateAudioRoute();
              String deviceName = intent.getStringExtra("name");
              if (deviceName == null) {
                deviceName = "";
              }
              WritableMap data = Arguments.createMap();
              data.putBoolean(
                "isPlugged",
                (intent.getIntExtra("state", 0) == 1) ? true : false
              );
              data.putBoolean(
                "hasMic",
                (intent.getIntExtra("microphone", 0) == 1) ? true : false
              );
              data.putString("deviceName", deviceName);
              sendEvent("WiredHeadset", data);
            } else {
              hasWiredHeadset = false;
            }
          }
        };
      ReactContext reactContext = getReactApplicationContext();
      if (reactContext != null) {
        reactContext.registerReceiver(wiredHeadsetReceiver, filter);
      } else {
        Log.d(TAG, "startWiredHeadsetEvent() reactContext is null");
      }
    }
  }

  private void stopWiredHeadsetEvent() {
    if (wiredHeadsetReceiver != null) {
      Log.d(TAG, "stopWiredHeadsetEvent()");
      this.unregisterReceiver(this.wiredHeadsetReceiver);
      wiredHeadsetReceiver = null;
    }
  }

  private void sendEvent(final String eventName, @Nullable WritableMap params) {
    try {
      ReactContext reactContext = getReactApplicationContext();
      if (reactContext != null && reactContext.hasActiveCatalystInstance()) {
        reactContext
          .getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
          .emit(eventName, params);
      } else {
        Log.e(
          TAG,
          "sendEvent(): reactContext is null or not having CatalystInstance yet."
        );
      }
    } catch (RuntimeException e) {
      Log.e(
        TAG,
        "sendEvent(): java.lang.RuntimeException: Trying to invoke JS before CatalystInstance has been set!"
      );
    }
  }

  @ReactMethod
  public void start(final boolean auto) {
    automatic = auto;
    if (!audioManagerActivated) {
      audioManagerActivated = true;

      Log.d(TAG, "start audioRouteManager");
      storeOriginalAudioSetup();
      startWiredHeadsetEvent();
      bluetoothManager.start();
      setSpeakerphoneOn(true);
      setMicrophoneMute(false);
      forceSpeakerOn = 0;
      hasWiredHeadset = hasWiredHeadset();
      defaultAudioDevice =
        (defaultSpeakerOn)
          ? AudioDevice.SPEAKER_PHONE
          : (hasEarpiece()) ? AudioDevice.EARPIECE : AudioDevice.SPEAKER_PHONE;
      userSelectedAudioDevice = AudioDevice.NONE;
      selectedAudioDevice = AudioDevice.NONE;
      audioDevices.clear();
      updateAudioRoute();
    }
  }

  public void stop() {
    stop();
  }

  @ReactMethod
  public void stop() {
    if (audioManagerActivated) {
      Log.d(TAG, "stop() AudioDeviceManager");
      stopWiredHeadsetEvent();
      setSpeakerphoneOn(false);
      setMicrophoneMute(false);
      forceSpeakerOn = 0;
      bluetoothManager.stop();
      restoreOriginalAudioSetup();
      audioManagerActivated = false;
    }
  }

  @ReactMethod
  public void setSpeakerphoneOn(final boolean enable) {
    if (enable != audioManager.isSpeakerphoneOn()) {
      Log.d(TAG, "setSpeakerphoneOn(): " + enable);
      audioManager.setSpeakerphoneOn(enable);
    }
  }

  // --- TODO (zxcpoiu): These two api name is really confusing. should be changed.
  /**
   * flag: Int
   * 0: use default action
   * 1: force speaker on
   * -1: force speaker off
   */
  @ReactMethod
  public void setForceSpeakerphoneOn(final int flag) {
    if (flag < -1 || flag > 1) {
      return;
    }
    Log.d(TAG, "setForceSpeakerphoneOn() flag: " + flag);
    forceSpeakerOn = flag;

    // --- will call updateAudioDeviceState()
    // --- Note: in some devices, it may not contains specified route thus will not be effected.
    if (flag == 1) {
      selectAudioDevice(AudioDevice.SPEAKER_PHONE);
    } else if (flag == -1) {
      selectAudioDevice(AudioDevice.EARPIECE); // --- use the most common earpiece to force `speaker off`
    } else {
      selectAudioDevice(AudioDevice.NONE); // --- NONE will follow default route, the default route of `video` call is speaker.
    }
  }

  // --- TODO (zxcpoiu): Implement api to let user choose audio devices

  @ReactMethod
  public void setMicrophoneMute(final boolean enable) {
    if (enable != audioManager.isMicrophoneMute()) {
      Log.d(TAG, "setMicrophoneMute(): " + enable);
      audioManager.setMicrophoneMute(enable);
    }
  }

  //  ===== Permission Start =====
  @ReactMethod
  public void checkRecordPermission(Promise promise) {
    Log.d(TAG, "RNAudioDeviceManager.checkRecordPermission(): enter");
    _checkRecordPermission();
    if (recordPermission.equals("unknow")) {
      Log.d(TAG, "RNAudioDeviceManager.checkRecordPermission(): failed");
      promise.reject(new Exception("checkRecordPermission failed"));
    } else {
      promise.resolve(recordPermission);
    }
  }

  private void _checkRecordPermission() {
    recordPermission = _checkPermission(permission.RECORD_AUDIO);
    Log.d(
      TAG,
      String.format(
        "RNAudioDeviceManager.checkRecordPermission(): recordPermission=%s",
        recordPermission
      )
    );
  }

  private String _checkPermission(String targetPermission) {
    try {
      ReactContext reactContext = getReactApplicationContext();
      if (
        ContextCompat.checkSelfPermission(reactContext, targetPermission) ==
        PackageManager.PERMISSION_GRANTED
      ) {
        return "granted";
      } else {
        return "denied";
      }
    } catch (Exception e) {
      Log.d(TAG, "_checkPermission() catch");
      return "denied";
    }
  }

  @ReactMethod
  public void requestRecordPermission(Promise promise) {
    Log.d(TAG, "RNAudioDeviceManager.requestRecordPermission(): enter");
    _checkRecordPermission();
    if (!recordPermission.equals("granted")) {
      _requestPermission(permission.RECORD_AUDIO, promise);
    } else {
      // --- already granted
      promise.resolve(recordPermission);
    }
  }

  @ReactMethod
  public void chooseAudioRoute(String audioRoute, Promise promise) {
    Log.d(
      TAG,
      "RNAudioDeviceManager.chooseAudioRoute(): user choose audioDevice = " +
      audioRoute
    );

    if (audioRoute.equals(AudioDevice.EARPIECE.name())) {
      selectAudioDevice(AudioDevice.EARPIECE);
    } else if (audioRoute.equals(AudioDevice.SPEAKER_PHONE.name())) {
      selectAudioDevice(AudioDevice.SPEAKER_PHONE);
    } else if (audioRoute.equals(AudioDevice.WIRED_HEADSET.name())) {
      selectAudioDevice(AudioDevice.WIRED_HEADSET);
    } else if (audioRoute.equals(AudioDevice.BLUETOOTH.name())) {
      selectAudioDevice(AudioDevice.BLUETOOTH);
    }
    promise.resolve(getAudioDeviceStatusMap());
  }

  private void _requestPermission(String targetPermission, Promise promise) {
    Activity currentActivity = getCurrentActivity();
    if (currentActivity == null) {
      Log.d(
        TAG,
        String.format(
          "RNAudioDeviceManager._requestPermission(): ReactContext doesn't hava any Activity attached when requesting %s",
          targetPermission
        )
      );
      promise.reject(
        new Exception("_requestPermission(): currentActivity is not attached")
      );
      return;
    }
    int requestPermissionCode = getRandomInteger(1, 65535);
    while (
      mRequestPermissionCodePromises.get(requestPermissionCode, null) != null
    ) {
      requestPermissionCode = getRandomInteger(1, 65535);
    }
    mRequestPermissionCodePromises.put(requestPermissionCode, promise);
    mRequestPermissionCodeTargetPermission.put(
      requestPermissionCode,
      targetPermission
    );
    ActivityCompat.requestPermissions(
      currentActivity,
      new String[] { targetPermission },
      requestPermissionCode
    );
  }

  private static int getRandomInteger(int min, int max) {
    if (min >= max) {
      throw new IllegalArgumentException("max must be greater than min");
    }
    Random random = new Random();
    return random.nextInt((max - min) + 1) + min;
  }

  protected static void onRequestPermissionsResult(
    int requestCode,
    String[] permissions,
    int[] grantResults
  ) {
    Log.d(TAG, "RNAudioDeviceManager.onRequestPermissionsResult(): enter");
    Promise promise = mRequestPermissionCodePromises.get(requestCode, null);
    String targetPermission = mRequestPermissionCodeTargetPermission.get(
      requestCode,
      null
    );
    mRequestPermissionCodePromises.delete(requestCode);
    mRequestPermissionCodeTargetPermission.delete(requestCode);
    if (promise != null && targetPermission != null) {
      Map<String, Integer> permissionResultMap = new HashMap<String, Integer>();

      for (int i = 0; i < permissions.length; i++) {
        permissionResultMap.put(permissions[i], grantResults[i]);
      }

      if (!permissionResultMap.containsKey(targetPermission)) {
        Log.wtf(
          TAG,
          String.format(
            "RNAudioDeviceManager.onRequestPermissionsResult(): requested permission %s but did not appear",
            targetPermission
          )
        );
        promise.reject(
          String.format("%s_PERMISSION_NOT_FOUND", targetPermission),
          String.format(
            "requested permission %s but did not appear",
            targetPermission
          )
        );
        return;
      }

      String _requestPermissionResult = "unknow";
      if (
        permissionResultMap.get(targetPermission) ==
        PackageManager.PERMISSION_GRANTED
      ) {
        _requestPermissionResult = "granted";
      } else {
        _requestPermissionResult = "denied";
      }

      if (targetPermission.equals(permission.RECORD_AUDIO)) {
        recordPermission = _requestPermissionResult;
      } else if (targetPermission.equals(permission.CAMERA)) {
        cameraPermission = _requestPermissionResult;
      }
      promise.resolve(_requestPermissionResult);
    } else {
      //super.onRequestPermissionsResult(requestCode, permissions, grantResults);
      Log.wtf(
        TAG,
        "RNAudioDeviceManager.onRequestPermissionsResult(): request code not found"
      );
      promise.reject(
        "PERMISSION_REQUEST_CODE_NOT_FOUND",
        "request code not found"
      );
    }
  }

  //  ===== Permission End =====

  private void updateAudioRoute() {
    if (!automatic) {
      return;
    }
    updateAudioDeviceState();
  }

  /** Changes selection of the currently active audio device. */
  private void setAudioDeviceInternal(AudioDevice device) {
    Log.d(TAG, "setAudioDeviceInternal(device=" + device + ")");
    if (!audioDevices.contains(device)) {
      Log.e(TAG, "specified audio device does not exist");
      return;
    }

    switch (device) {
      case SPEAKER_PHONE:
        setSpeakerphoneOn(true);
        break;
      case EARPIECE:
        setSpeakerphoneOn(false);
        break;
      case WIRED_HEADSET:
        setSpeakerphoneOn(false);
        break;
      case BLUETOOTH:
        setSpeakerphoneOn(false);
        break;
      default:
        Log.e(TAG, "Invalid audio device selection");
        break;
    }
    selectedAudioDevice = device;
  }

  /**
   * Changes default audio device.
   * TODO(henrika): add usage of this method in the AppRTCMobile client.
   */
  public void setDefaultAudioDevice(AudioDevice defaultDevice) {
    switch (defaultDevice) {
      case SPEAKER_PHONE:
        defaultAudioDevice = defaultDevice;
        break;
      case EARPIECE:
        if (hasEarpiece()) {
          defaultAudioDevice = defaultDevice;
        } else {
          defaultAudioDevice = AudioDevice.SPEAKER_PHONE;
        }
        break;
      default:
        Log.e(TAG, "Invalid default audio device selection");
        break;
    }
    Log.d(TAG, "setDefaultAudioDevice(device=" + defaultAudioDevice + ")");
    updateAudioDeviceState();
  }

  /** Changes selection of the currently active audio device. */
  public void selectAudioDevice(AudioDevice device) {
    if (device != AudioDevice.NONE && !audioDevices.contains(device)) {
      Log.e(
        TAG,
        "selectAudioDevice() Can not select " +
        device +
        " from available " +
        audioDevices
      );
      return;
    }
    userSelectedAudioDevice = device;
    updateAudioDeviceState();
  }

  /** Returns current set of available/selectable audio devices. */
  public Set<AudioDevice> getAudioDevices() {
    return Collections.unmodifiableSet(new HashSet<>(audioDevices));
  }

  /** Returns the currently selected audio device. */
  public AudioDevice getSelectedAudioDevice() {
    return selectedAudioDevice;
  }

  /** Helper method for receiver registration. */
  private void registerReceiver(
    BroadcastReceiver receiver,
    IntentFilter filter
  ) {
    getReactApplicationContext().registerReceiver(receiver, filter);
  }

  /** Helper method for unregistration of an existing receiver. */
  private void unregisterReceiver(final BroadcastReceiver receiver) {
    final ReactContext reactContext = this.getReactApplicationContext();
    if (reactContext != null) {
      try {
        reactContext.unregisterReceiver(receiver);
      } catch (final Exception e) {
        Log.d(TAG, "unregisterReceiver() failed");
      }
    } else {
      Log.d(TAG, "unregisterReceiver() reactContext is null");
    }
  }

  /** Gets the current earpiece state. */
  private boolean hasEarpiece() {
    return getReactApplicationContext()
      .getPackageManager()
      .hasSystemFeature(PackageManager.FEATURE_TELEPHONY);
  }

  /**
   * Checks whether a wired headset is connected or not.
   * This is not a valid indication that audio playback is actually over
   * the wired headset as audio routing depends on other conditions. We
   * only use it as an early indicator (during initialization) of an attached
   * wired headset.
   */
  @Deprecated
  private boolean hasWiredHeadset() {
    if (Build.VERSION.SDK_INT < Build.VERSION_CODES.M) {
      return audioManager.isWiredHeadsetOn();
    } else {
      final AudioDeviceInfo[] devices = audioManager.getDevices(
        AudioManager.GET_DEVICES_ALL
      );
      for (AudioDeviceInfo device : devices) {
        final int type = device.getType();
        if (type == AudioDeviceInfo.TYPE_WIRED_HEADSET) {
          Log.d(TAG, "hasWiredHeadset: found wired headset");
          return true;
        } else if (type == AudioDeviceInfo.TYPE_USB_DEVICE) {
          Log.d(TAG, "hasWiredHeadset: found USB audio device");
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Updates list of possible audio devices and make new device selection.
   */
  public void updateAudioDeviceState() {
    Log.d(
      TAG,
      "--- updateAudioDeviceState: " +
      "wired headset=" +
      hasWiredHeadset +
      ", " +
      "BT state=" +
      bluetoothManager.getState()
    );
    Log.d(
      TAG,
      "Device status: " +
      "available=" +
      audioDevices +
      ", " +
      "selected=" +
      selectedAudioDevice +
      ", " +
      "user selected=" +
      userSelectedAudioDevice
    );

    // Check if any Bluetooth headset is connected. The internal BT state will
    // change accordingly.
    // TODO(henrika): perhaps wrap required state into BT manager.
    if (
      bluetoothManager.getState() == BluetoothManager.State.HEADSET_AVAILABLE ||
      bluetoothManager.getState() ==
      BluetoothManager.State.HEADSET_UNAVAILABLE ||
      bluetoothManager.getState() == BluetoothManager.State.SCO_DISCONNECTING
    ) {
      bluetoothManager.updateDevice();
    }

    // Update the set of available audio devices.
    Set<AudioDevice> newAudioDevices = new HashSet<>();

    // always assume device has speaker phone
    newAudioDevices.add(AudioDevice.SPEAKER_PHONE);

    if (
      bluetoothManager.getState() == BluetoothManager.State.SCO_CONNECTED ||
      bluetoothManager.getState() == BluetoothManager.State.SCO_CONNECTING ||
      bluetoothManager.getState() == BluetoothManager.State.HEADSET_AVAILABLE
    ) {
      newAudioDevices.add(AudioDevice.BLUETOOTH);
    }

    if (hasWiredHeadset) {
      newAudioDevices.add(AudioDevice.WIRED_HEADSET);
    }

    if (hasEarpiece()) {
      newAudioDevices.add(AudioDevice.EARPIECE);
    }

    // --- check whether user selected audio device is available
    if (
      userSelectedAudioDevice != null &&
      userSelectedAudioDevice != AudioDevice.NONE &&
      !newAudioDevices.contains(userSelectedAudioDevice)
    ) {
      userSelectedAudioDevice = AudioDevice.NONE;
    }

    // Store state which is set to true if the device list has changed.
    boolean audioDeviceSetUpdated = !audioDevices.equals(newAudioDevices);
    // Update the existing audio device set.
    audioDevices = newAudioDevices;

    AudioDevice newAudioDevice = getPreferredAudioDevice();

    // --- stop bluetooth if needed
    if (
      selectedAudioDevice == AudioDevice.BLUETOOTH &&
      newAudioDevice != AudioDevice.BLUETOOTH &&
      (
        bluetoothManager.getState() == BluetoothManager.State.SCO_CONNECTED ||
        bluetoothManager.getState() == BluetoothManager.State.SCO_CONNECTING
      )
    ) {
      bluetoothManager.stopScoAudio();
      bluetoothManager.updateDevice();
    }

    // --- start bluetooth if needed
    if (
      selectedAudioDevice != AudioDevice.BLUETOOTH &&
      newAudioDevice == AudioDevice.BLUETOOTH &&
      bluetoothManager.getState() == BluetoothManager.State.HEADSET_AVAILABLE
    ) {
      // Attempt to start Bluetooth SCO audio (takes a few second to start).
      if (!bluetoothManager.startScoAudio()) {
        // Remove BLUETOOTH from list of available devices since SCO failed.
        audioDevices.remove(AudioDevice.BLUETOOTH);
        audioDeviceSetUpdated = true;
        if (userSelectedAudioDevice == AudioDevice.BLUETOOTH) {
          userSelectedAudioDevice = AudioDevice.NONE;
        }
        newAudioDevice = getPreferredAudioDevice();
      }
    }

    if (
      newAudioDevice == AudioDevice.BLUETOOTH &&
      bluetoothManager.getState() != BluetoothManager.State.SCO_CONNECTED
    ) {
      newAudioDevice = getPreferredAudioDevice(true); // --- skip bluetooth
    }

    // Switch to new device but only if there has been any changes.
    if (newAudioDevice != selectedAudioDevice || audioDeviceSetUpdated) {
      // Do the required device switch.
      setAudioDeviceInternal(newAudioDevice);
      Log.d(
        TAG,
        "New device status: " +
        "available=" +
        audioDevices +
        ", " +
        "selected=" +
        newAudioDevice
      );
      /*
            if (audioManagerEvents != null) {
                // Notify a listening client that audio device has been changed.
                audioManagerEvents.onAudioDeviceChanged(selectedAudioDevice, audioDevices);
            }
            */
      sendEvent("onAudioDeviceChanged", getAudioDeviceStatusMap());
    }
    Log.d(TAG, "--- updateAudioDeviceState done");
  }

  private WritableMap getAudioDeviceStatusMap() {
    WritableMap data = Arguments.createMap();
    String audioDevicesJson = "[";
    for (AudioDevice s : audioDevices) {
      audioDevicesJson += "\"" + s.name() + "\",";
    }

    // --- strip the last `,`
    if (audioDevicesJson.length() > 1) {
      audioDevicesJson =
        audioDevicesJson.substring(0, audioDevicesJson.length() - 1);
    }
    audioDevicesJson += "]";

    data.putString("availableAudioDeviceList", audioDevicesJson);
    data.putString(
      "selectedAudioDevice",
      (selectedAudioDevice == null) ? "" : selectedAudioDevice.name()
    );

    return data;
  }

  private AudioDevice getPreferredAudioDevice() {
    return getPreferredAudioDevice(false);
  }

  private AudioDevice getPreferredAudioDevice(boolean skipBluetooth) {
    final AudioDevice newAudioDevice;

    if (
      userSelectedAudioDevice != null &&
      userSelectedAudioDevice != AudioDevice.NONE
    ) {
      newAudioDevice = userSelectedAudioDevice;
    } else if (!skipBluetooth && audioDevices.contains(AudioDevice.BLUETOOTH)) {
      // If a Bluetooth is connected, then it should be used as output audio
      // device. Note that it is not sufficient that a headset is available;
      // an active SCO channel must also be up and running.
      newAudioDevice = AudioDevice.BLUETOOTH;
    } else if (audioDevices.contains(AudioDevice.WIRED_HEADSET)) {
      // If a wired headset is connected, but Bluetooth is not, then wired headset is used as
      // audio device.
      newAudioDevice = AudioDevice.WIRED_HEADSET;
    } else if (audioDevices.contains(defaultAudioDevice)) {
      newAudioDevice = defaultAudioDevice;
    } else {
      newAudioDevice = AudioDevice.SPEAKER_PHONE;
    }

    return newAudioDevice;
  }
}
